!2345678911234567892123456789312345678941234567895123456789612345678971234567898
!        0         0         0         0         0         0         0         0
!                   
! Module: merge_mod
!
! Authors:  Andy Read - XMM-SSC-MPE
!          
! Description: 
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module merge_mod

  use cal
  use oal
  use dal
  use param  
  use errorHandling
  
  implicit none 
  
  private 

  public merge

contains 
  
  !---------------------------------------------------------------------
  subroutine merge
  !---------------------------------------------------------------------

  type(DataSetT)                :: set1, set2
  character(len=8)              :: type

!******************************************************************************
!	START 
!
!-----------------------------------------------------------------------
!       GET TYPES - CALL MAIN ROUTINE
!-----------------------------------------------------------------------
    set1 = dataset(stringParameter("set1"),READ)
    set2 = dataset(stringParameter("set2"),READ)
    if (hasBlock(set1,"EVENTS") .and. hasBlock(set2,"EVENTS")) then 
      type="events"
      call ev(type)
    else if (hasBlock(set1,"ATTHK") .and. hasBlock(set2,"ATTHK")) then
      type="attitude"
      call att(type)
    else if (hasBlock(set1,"ORBIT") .and. hasBlock(set2,"ORBIT")) then
      type="orbit"
      call orb(type)
    else
      call error("filesNonCompatible","Input files are non-compatible (File types different)")
    endif

    end subroutine merge
!******************************************************************************
    subroutine ev (type)

  implicit none

  type(DataSetT)                :: set1, set2, outset

  character(len=40)             :: attr1, attr2
  character(len=8)              :: type

  real(double)                  :: ra, dec

  integer                       :: i, Nexpos

  logical			:: reverse, firsttab, sameinstrument
                                          
  integer, parameter            :: Ncolumns = 17
  character(len=8), parameter   :: tabname = "EVENTS  "
  character(len=8), dimension(17), parameter :: colname = &
    (/"TIME    ","RAWX    ","RAWY    ","DETX    ","DETY    ","X       ",&
      "Y       ","PHA     ","PI      ","FLAG    ","PATTERN ","PAT_ID  ",&
      "PAT_SEQ ","PAT_TYP ","PAT_IND ","PAT_ORI ","CCDNR   "/)
  character(len=3), dimension(17), parameter :: coltype = &
    (/"R64","I16","I16","I16","I16","I32","I32","I16","I16","I32","I08","I16","I08","I16","I16",&
      "I16","I08"/)

  integer, parameter            :: NEcolumns = 8
  character(len=8), dimension(12), parameter :: Etabname = &
    (/"EXPOSU01","EXPOSU02","EXPOSU03","EXPOSU04","EXPOSU05","EXPOSU06",&
      "EXPOSU07","EXPOSU08","EXPOSU09","EXPOSU10","EXPOSU11","EXPOSU12"/)
  character(len=8), dimension(8), parameter :: Ecolname = &
    (/"TIME    ","TIMEDEL ","FRACEXP ","NDSCLIN1","NDSCLIN2","NDSCLIN3",&
      "NDSCLIN4","FRAMELIM"/)
  character(len=3), dimension(8), parameter :: Ecoltype = &
    (/"R64","R32","R32","I16","I16","I16",&
      "I16","I32"/)
  
  integer, parameter            :: NGTIcolumns = 2
  character(len=8), dimension(12), parameter :: GTItabname = &
    (/"STDGTI01","STDGTI02","STDGTI03","STDGTI04","STDGTI05","STDGTI06",&
      "STDGTI07","STDGTI08","STDGTI09","STDGTI10","STDGTI11","STDGTI12"/)
  character(len=8), dimension(2), parameter :: GTIcolname = &
    (/"START   ","STOP    "/)
  character(len=3), dimension(2), parameter :: GTIcoltype = &
    (/"R64","R64"/)
  
!------------------------------------------------
!	GET PARAMETERS 	
!------------------------------------------------
    set1 = dataset(stringParameter("set1"),READ)
    set2 = dataset(stringParameter("set2"),READ)
    ra = realParameter("ra")
    dec = realParameter("dec")
!------------------------------------------
!     Ev Files consistent?
!------------------------------------------
    write (errstr,*) "Checking files: ",trim(stringParameter("set1"))," and ",trim(stringParameter("set2"))
    call message(AppMsg,Verbose)
    attr1 = stringAttribute(set1,"INSTRUME")
    attr2 = stringAttribute(set2,"INSTRUME")
    if (trim(attr1) /= "EPN" .and. trim(attr1) /= "EMOS1" .and. trim(attr1) /= "EMOS2") then 
       call error("badInstrument","Instrument not supported")
    endif
    if (trim(attr2) /= "EPN" .and. trim(attr2) /= "EMOS1" .and. trim(attr2) /= "EMOS2") then 
       call error("badInstrument","Instrument not supported")
    endif
    sameinstrument=.TRUE.
    if (trim(attr1) /= trim(attr2)) then  
      sameinstrument=.FALSE.
!      call error("filesNonCompatible","Input files are non-compatible (INSTRUME)")
      call warning("filesNonCompatible","Input files are from different instruments")
    else
      if (trim(attr1) == "EPN") then 		!both are epn
        Nexpos = 12        
      else					!both mos        
        Nexpos = 7
      endif
    endif
    write (errstr,*) "Instruments: ",trim(attr1),", ",trim(attr2)
    call message(AppMsg,Verbose)
    attr1 = stringAttribute(set1,"FILTER")
    attr2 = stringAttribute(set2,"FILTER")
    if (trim(attr1) /= trim(attr2)) then  
!      call error("filesNonCompatible","Input files are non-compatible (FILTER)")
      call warning("filesNonCompatible","Input files use different filter")
    endif
    write (errstr,*) "Filter: ",trim(attr1)
    call message(AppMsg,Verbose)
    attr1 = stringAttribute(set1,"DATAMODE")
    attr2 = stringAttribute(set2,"DATAMODE")
    if (trim(attr1) /= trim(attr2)) then  
      call error("filesNonCompatible","Input files are non-compatible (DATAMODE)")
    endif
    write (errstr,*) "DATAMODE: ",trim(attr1)
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"Files consistent")
!------------------------------------------
!     Merge event tables
!------------------------------------------
    firsttab = .TRUE.
    call mergetab(set1,set2,type,tabname,Ncolumns,colname,coltype,&
                  reverse,firsttab,outset)
!------------------------------------------
!     Convert X,Ys
!------------------------------------------
    if (reverse) call XY(set2,set1,ra,dec,outset,reverse)
    if (.not.reverse) call XY(set1,set2,ra,dec,outset,reverse)
!------------------------------------------
!	Merge exposure extensions     
!------------------------------------------
    firsttab = .FALSE.
    if (sameinstrument) then 
      do i = 1, Nexpos
        call mergetab(set1,set2,type,Etabname(i),NEcolumns,Ecolname,Ecoltype,&
                      reverse,firsttab,outset)
      enddo
    else
      call warning("noExposureMerge","Input files are from different instruments - EXPOSURE extensions not merged")
    endif
!------------------------------------------
!	Merge GTI extensions     
!------------------------------------------
    firsttab = .FALSE.
    if (sameinstrument) then 
      do i = 1, Nexpos
        call mergetab(set1,set2,type,GTItabname(i),NGTIcolumns,GTIcolname,GTIcoltype,&
                      reverse,firsttab,outset)
      enddo
    else
      call warning("noGTIMerge","Input files are from different instruments - GTI extensions not merged")
    endif

    call release(outset)

    end subroutine ev
!******************************************************************************

    subroutine att(type)

  implicit none

  type(DataSetT)                :: set1, set2, outset

  character(len=40)             :: attr1, attr2
  character(len=8)              :: type

  logical			:: reverse, firsttab
                                          
  integer, parameter            :: Ncolumns = 10
  character(len=8), parameter   :: tabname = "ATTHK   "
  character(len=8), dimension(10), parameter :: colname = &
    (/"TIME    ","AHFRA   ","AHFDEC  ","AHFPA   ","OMRA    ","OMDEC   ",&
      "OMPA    ","DAHFNOM ","DOMNOM  ","DAHFOM  "/)
  character(len=3), dimension(10), parameter :: coltype = &
    (/"R64","R64","R64","R64","R64","R64","R64","R64","R64","R64"/)

!------------------------------------------------
!	GET PARAMETERS 	
!------------------------------------------------
    set1 = dataset(stringParameter("set1"),READ)
    set2 = dataset(stringParameter("set2"),READ)
!------------------------------------------
!     Files consistent?
!------------------------------------------
    write (errstr,*) "Checking files: ",trim(stringParameter("set1"))," and ",trim(stringParameter("set2"))
    call message(AppMsg,Verbose)
    attr1 = stringAttribute(set1,"TELESCOP")
    attr2 = stringAttribute(set2,"TELESCOP")
    if (trim(attr1) /= trim(attr2)) then  
      call error("filesNonCompatible","Input files are non-compatible (TELESCOP)")
    endif
!    write (errstr,*) "Telescope: ",trim(attr1)
!    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"Files consistent")
!------------------------------------------
!     Merge atthk extensions 
!------------------------------------------
    firsttab = .TRUE.
    call mergetab(set1,set2,type,tabname,Ncolumns,colname,coltype,&
                  reverse,firsttab,outset)
!------------------------------------------
!     
!------------------------------------------

    call release(outset)

    end subroutine att

!******************************************************************************
    subroutine headatt(outset,set1,set2,reverse)

    implicit none

  type(DataSetT), intent(inout)      :: outset 
  type(DataSetT)                     :: set1, set2

  logical                            :: reverse

  real(double)			     :: value

!--------------------------------------------------------------
!	Change att header
!--------------------------------------------------------------  
    call setAttribute(outset,"DATE-END",&
        trim(stringAttribute(set2,"DATE-END")),"Date observations ended")

    value =( ( int16Attribute(set1,"NGAHF")*real64Attribute(set1,"AAHFRA") ) +&
             ( int16Attribute(set2,"NGAHF")*real64Attribute(set2,"AAHFRA") ) )/&
             (int16Attribute(set1,"NGAHF")+int16Attribute(set2,"NGAHF"))

!write(*,*) value !So, possible!

    end subroutine headatt
!******************************************************************************
!******************************************************************************
    subroutine mergetab(set1,set2,type,tabname,Ncolumns,colname,coltype,&
                        reverse,firsttab,outset)

    implicit none

  type(DataSetT), intent(inout) :: set1, set2 
  type(DataSetT)                :: outset
  type(tableT)                  :: tab1, tab2, tab
  type(ColumnT)                 :: col1, col2 , col

  integer                       :: n, i, Ncolumns

  real(double), dimension(:), pointer   :: pR64_1, pR64_2, pR64
  real(single), dimension(:), pointer   :: pR32_1, pR32_2, pR32
  integer(int32), dimension(:), pointer :: pI32_1, pI32_2, pI32
  integer(int16), dimension(:), pointer :: pI16_1, pI16_2, pI16
  integer(int8), dimension(:), pointer  :: pI8_1, pI8_2, pI8

  real(double)                  :: start, end, start1, start2, end1, end2
                                          
  real, parameter               :: radian = 57.29577951E0, &
                                   skypixsize = 1.3888888888E-5, &!0.05arcsec, 
                                   pi = 3.1415926536E0

  logical                       :: reverse, firsttab

  character(len=8)              :: type

  character(len=8)              :: tabname
  character(len=8), dimension(Ncolumns) :: colname 
  character(len=3), dimension(Ncolumns) :: coltype


    if (trim(type) == "events") then 
      write (errstr,*) "EVENTS: Merging ",trim(tabname)," extensions"
      call message(AppMsg,Verbose)
    elseif (trim(type) == "attitude") then
      write (errstr,*) "ATTITUDE: Merging ",trim(tabname)," extensions"
      call message(AppMsg,Verbose)
    elseif (trim(type) == "hk") then
      write (errstr,*) "HK: Merging ",trim(tabname)," extensions"
      call message(AppMsg,Verbose)
    elseif (trim(type) == "orbit") then
      write (errstr,*) "ORBIT: Merging ",trim(tabname)," extensions"
      call message(AppMsg,Verbose)
    endif
!------------------------------------------------
!	Get times and `reverse' (firsttab)
!------------------------------------------------
    if (.not.hasBlock(set1,tabname) .or. .not.hasBlock(set2,tabname)) then
      write (errstr,*) tabname," extension does not exist in one or both files - not merging"
      call warning("noExtension")
    else 
      tab1 = table(set1,tabname)
      tab2 = table(set2,tabname)

      if (firsttab) then 
        if (hasColumn(tab1,"TIME") .and. hasColumn(tab2,"TIME")) then
          col1 = column(tab1,"TIME",READ)
          col2 = column(tab2,"TIME",READ)
          pR64_1 => real64Data(col1)
          pR64_2 => real64Data(col2)
      
          end = -9999999999.9
          start = 9999999999.9
          do n = 0, size(pR64_1)-1
            if (pR64_1(n) > end) end = pR64_1(n)
            if (pR64_1(n) < start) start = pR64_1(n)
          enddo
          write (errstr,*) trim(stringParameter("set1")),": ",size(pR64_1)," entries"
          call message(AppMsg,Verbose)
          write (errstr,*) trim(stringParameter("set1")),": ",start," to ",end," s"
          call message(AppMsg,Verbose)
          start1 = start
          end1 = end
        
          end = -9999999999.9
          start = 9999999999.9
          do n = 0, size(pR64_2)-1
            if (pR64_2(n) > end) end = pR64_2(n)
            if (pR64_2(n) < start) start = pR64_2(n)
          enddo
          write (errstr,*) trim(stringParameter("set2")),": ",size(pR64_2)," entries"
          call message(AppMsg,Verbose)
          write (errstr,*) trim(stringParameter("set2")),": ",start," to ",end," s"
          call message(AppMsg,Verbose)
          start2 = start
          end2 = end
          call release(col1)
          call release(col2)
      
          if (start2 > start1) then 
            reverse = .FALSE.
            if (start2 < end1) then 
              call message(AppMsg,Verbose,"Files overlap in time")
            endif  
          else
            reverse = .TRUE.
            call message(AppMsg,Verbose,"Input files in wrong time order - reversing")
            if (start1 < end2) then 
              call message(AppMsg,Verbose,"Files overlap in time")
            endif  
          endif
        else
          write (errstr,*) "TIME column missing in one file - cannot establish order"
          call warning("noTimeColumn")
          reverse = .FALSE.
        endif      
      endif 	!firsttab
    
      if (numberOfColumns(tab1) /= numberOfColumns(tab2)) then 
        write (errstr,*) "No of columns:",trim(tabname),"(",numberOfColumns(tab1),",",numberOfColumns(tab2),")"
        call message(AppMsg,Verbose)
      else
        write (errstr,*) numberOfColumns(tab1)," columns in each ",trim(tabname)," extension"
        call message(AppMsg,Verbose)
      endif
  
      if (firsttab) then 
        if (.not.reverse) call copyDataSet(stringParameter("set1"),stringParameter("outset"))
        if (reverse) call copyDataSet(stringParameter("set2"),stringParameter("outset"))
        outset = dataset(stringParameter("outset"),MODIFY)
      endif

      tab = table(outset,tabname)
      if (.not.reverse) call insertRows(tab,numberOfRows(tab1),numberOfRows(tab2))
      if (reverse) call insertRows(tab,numberOfRows(tab2),numberOfRows(tab1))

      do i = 1, Ncolumns
  
        if (hasColumn(tab1,colname(i)) .and. hasColumn(tab2,colname(i))) then
          col = column(tab,colname(i),MODIFY)
          if (.not.reverse) then 
            col1 = column(tab1,colname(i),READ)
            col2 = column(tab2,colname(i),READ)
          else
            col2 = column(tab1,colname(i),READ)
            col1 = column(tab2,colname(i),READ)
          endif
          if (coltype(i) == "R64") then 
            pR64_1 => real64Data(col1)
            pR64_2 => real64Data(col2)
            pR64 =>  real64Data(col)
              do n = 0, size(pR64_2)-1
                pR64(size(pR64_1)+n) = pR64_2(n)
              enddo
          elseif (coltype(i) == "R32") then
            pR32_1 => real32Data(col1)
            pR32_2 => real32Data(col2)
            pR32 =>  real32Data(col)
              do n = 0, size(pR32_2)-1
                pR32(size(pR32_1)+n) = pR32_2(n)
              enddo
          elseif (coltype(i) == "I32") then
            pI32_1 => int32Data(col1)
            pI32_2 => int32Data(col2)
            pI32 =>  int32Data(col)
              do n = 0, size(pI32_2)-1
                pI32(size(pI32_1)+n) = pI32_2(n)
              enddo
          elseif (coltype(i) == "I16") then
            pI16_1 => int16Data(col1)
            pI16_2 => int16Data(col2)
            pI16 =>  int16Data(col)
              do n = 0, size(pI16_2)-1
                pI16(size(pI16_1)+n) = pI16_2(n)
              enddo
          elseif (coltype(i) == "I08") then
            pI8_1 => int8Data(col1)
            pI8_2 => int8Data(col2)
            pI8 =>  int8Data(col)
              do n = 0, size(pI8_2)-1
                pI8(size(pI8_1)+n) = pI8_2(n)
              enddo
  !        elseif (coltype(i) == "Boo") then 
  !          p_col1 => booleanData(col1)
  !          p_col2 => booleanData(col2)
  !          p_col =>  booleanData(col)
          endif
          call release(col1)
          call release(col2)
        else
          if (.not.hasColumn(tab1,colname(i)) .and. .not.hasColumn(tab2,colname(i))) then
          else
            write (errstr,*) "Column ",colname(i)," missing in one file - not merging"
            call warning("noColumn")
            if ((.not.reverse .and. .not.hasColumn(tab2,colname(i))) .or. &
                  (reverse .and. .not.hasColumn(tab1,colname(i)))) then            
              call deleteColumn(tab,colname(i))
            endif
          endif
        endif
      enddo

      call release(tab)
      call release(tab1)
      call release(tab2)
  
      if (trim(type) == "events") then 
  !      call headev
      elseif (trim(type) == "attitude") then
        call headatt(outset,set1,set2,reverse)
      elseif (trim(type) == "hk") then
  !      call headhk
      elseif (trim(type) == "orbit") then
  !      call headorbit
      endif
  
    endif


  end subroutine mergetab
!******************************************************************************
    subroutine XY(set1,set2,outranom,outdecnom,outset,reverse)	!set1/2 ordered

    implicit none

  type(DataSetT), intent(inout) :: set1, set2 
  type(DatasetT)                :: outset
  type(tableT)                  :: tab1, tab2, tab 
  type(ColumnT)                 :: X1, Y1, X2, Y2, X, Y

  integer                       :: n, value, Nrows1, Xmin, Xmax, Ymin, Ymax

  integer(int32), dimension(:), pointer :: p_X1, p_Y1, p_X2, P_Y2, P_X, P_Y

  real(double)                  :: ra, dec, relx, rely, inranom, indecnom, &
				   outranom, outdecnom, soutdecnom, &
                                   coutdecnom, scaleX, scaleY, imsize, ra1, &
                                   ra2, dec1, dec2, offset

  real, parameter               :: radian = 57.29577951E0, &
                                   skypixsize = 1.3888888888E-5, &!0.05arcsec, 
                                   pi = 3.1415926536E0
  logical 			:: reverse

!------------------------
!	Open tables
!------------------------
    tab = table(outset,"EVENTS")
    if (hasColumn(tab,"X") .and. hasColumn(tab,"Y")) then 
      X = column(tab,"X",MODIFY)
      Y = column(tab,"Y",MODIFY)
      P_X => int32Data(X)
      P_Y => int32Data(Y)
      tab1 = table(set1,"EVENTS")
      tab2 = table(set2,"EVENTS")
      X1 = column(tab1,"X",READ)
      X2 = column(tab2,"X",READ)
      Y1 = column(tab1,"Y",READ)
      Y2 = column(tab2,"Y",READ)
      Xmin = 99999999
      Ymin = 99999999
      Xmax = -99999999
      Ymax = -99999999
      imsize = realParameter("imagesize")
      imsize = 72000.0 * imsize             !in 0.05arcs
      write(errstr,*) "Image size [in 0.05 arcs.] : ",imsize
      call message(AppMsg,Verbose)
!------------------------------------------------------
!	Get average pointing or user pointing and offset
!------------------------------------------------------
      ra1 = real64Attribute(X1,"TCRVL")/radian
      ra2 = real64Attribute(X2,"TCRVL")/radian
      dec1 = real64Attribute(Y1,"TCRVL")/radian
      dec2 = real64Attribute(Y2,"TCRVL")/radian
      if (.not. booleanParameter("withuserradec")) then  
        outranom = radian*((ra1+ra2)/2.0)
        outdecnom = radian*((dec1+dec2)/2.0)
        write (errstr,*) "Mean pointing will be used: ",outranom,", ",outdecnom
        call message(AppMsg, Verbose)
      else
        write (errstr,*) "User-input pointing will be used: ",outranom,", ",outdecnom
        call message(AppMsg, Verbose)
      endif
      outranom = outranom/radian
      soutdecnom = dsin(outdecnom/radian)
      coutdecnom = dcos(outdecnom/radian)
      offset = (dcos(ra1)*dcos(dec1)*dcos(ra2)*dcos(dec2)) + &
               (dsin(ra1)*dcos(dec1)*dsin(ra2)*dcos(dec2)) + &
               (dsin(dec1)*dsin(dec2)) 
      offset = radian * abs(dacos(offset))
      write (errstr,*) "Pointing difference: ",offset," degrees"
      call message(AppMsg,Verbose)
      if (offset > 5.0 .and. offset <= 10.0) then 
call warning("largePointingOffset","Pointings differ by more than 5 degrees - reprojection errors approaching arcminute scales")
      else if (offset > 10.0) then 
call warning("largePointingOffset","Pointings differ by more than 10 degrees - reprojection errors approaching 0.1 deg scales")
      endif      
  !--------------------
  !	1st set
  !--------------------
      p_X1 => int32Data(X1)
      p_Y1 => int32Data(Y1)
      inranom = real64Attribute(X1,"TCRVL")/radian
      indecnom = real64Attribute(Y1,"TCRVL")/radian
      scaleX = -1 * real64Attribute(X1,"TCDLT")/radian	!rad/pix
      scaleY = real64Attribute(Y1,"TCDLT")/radian		!rad/pix
      Nrows1 = numberOfRows(tab1)
  
      do n = 0, Nrows1 - 1

!  if (n==0 .or. n==2) write (*,*) "p_X1(n),p_Y1(n)",p_X1(n),p_Y1(n)
        relx = (p_X1(n) - int32Attribute(X1,"TCRPX")) * scaleX	!radians
        rely = - (p_Y1(n) - int32Attribute(Y1,"TCRPX")) * scaleY	!radians
!  if (n==0 .or. n==2) write (*,*) "relx,rely",relx*radian,rely*radian
        !-------------------------
        !    ANTISPHERE gets ra/dec of point
        !-------------------------
        if (n==0) call message(AppMsg,Verbose,"Projecting 1st dataset X/Y coordinates")
        call antisp(relx,rely,inranom,indecnom,ra,dec)
!  if (n==0 .or. n==2) write (*,*) "ra,dec",ra*radian,dec*radian
        !-------------------------
        !    SPHERE
        !-------------------------
        if (n==0) call message(AppMsg,Verbose,"Back-projecting 1st dataset X/Y coordinates")
        call sphere(ra,dec,outranom,soutdecnom,coutdecnom,relx,rely)
!  if (n==0 .or. n==2) write (*,*) "relx,rely",relx*radian,rely*radian
        p_X(n) = imsize + 1 - (relx/scaleX)	!placed into outset
        p_Y(n) = imsize + 1 + (rely/scaleY)
!  if (n==0 .or. n==2) write (*,*) "p_X(n),p_Y(n)",p_X(n),p_Y(n)
        if (p_X(n) < Xmin) Xmin = p_X(n)
        if (p_Y(n) < Ymin) Ymin = p_Y(n)
        if (p_X(n) > Xmax) Xmax = p_X(n)
        if (p_Y(n) > Ymax) Ymax = p_Y(n)
  
      enddo
!write (*,*) "Xmin,Ymin,Xmax,Ymax"
!write (*,*) Xmin,Ymin,Xmax,Ymax
      call release(X1)
      call release(Y1)
      call release(tab1)
  !--------------------
  !	2nd set
  !--------------------
      p_X2 => int32Data(X2)
      p_Y2 => int32Data(Y2)
      inranom = real64Attribute(X2,"TCRVL")/radian
      indecnom = real64Attribute(Y2,"TCRVL")/radian
      scaleX = -1 * real64Attribute(X2,"TCDLT")/radian	!rad/pix
      scaleY = real64Attribute(Y2,"TCDLT")/radian		!rad/pix
  
      do n = 0, numberOfRows(tab2)-1
!  if (n==0 .or. n==2) write (*,*) "p_X2(n),p_Y2(n)",p_X2(n),p_Y2(n)
  
        relx = (p_X2(n) - int32Attribute(X2,"TCRPX")) * scaleX	!radians
        rely = - (p_Y2(n) - int32Attribute(Y2,"TCRPX")) * scaleY	!radians
!  if (n==0 .or. n==2) write (*,*) "relx,rely",relx*radian,rely*radian
        !-------------------------
        !    ANTISPHERE gets ra/dec of point
        !-------------------------
        if (n==0) call message(AppMsg,Verbose,"Projecting 2nd dataset X/Y coordinates")
        call antisp(relx,rely,inranom,indecnom,ra,dec)
!  if (n==0 .or. n==2) write (*,*) "ra,dec",ra*radian,dec*radian
        !-------------------------
        !    SPHERE
        !-------------------------
        if (n==0) call message(AppMsg,Verbose,"Back-projecting 2nd dataset X/Y coordinates")
        call sphere(ra,dec,outranom,soutdecnom,coutdecnom,relx,rely)
!  if (n==0 .or. n==2) write (*,*) "relx,rely",relx*radian,rely*radian
        p_X(Nrows1 + n) = imsize + 1 - (relx/scaleX)
        p_Y(Nrows1 + n) = imsize + 1 + (rely/scaleY)
!  if (n==0 .or. n==2) write (*,*) "p_X('n'),p_Y('n')",p_X(Nrows1 + n),p_Y(Nrows1 + n)
        if (p_X(Nrows1 + n) < Xmin) Xmin = p_X(Nrows1 + n)
        if (p_Y(Nrows1 + n) < Ymin) Ymin = p_Y(Nrows1 + n)
        if (p_X(Nrows1 + n) > Xmax) Xmax = p_X(Nrows1 + n)
        if (p_Y(Nrows1 + n) > Ymax) Ymax = p_Y(Nrows1 + n)
  
      enddo
!write (*,*) "Xmin,Ymin,Xmax,Ymax"
!write (*,*) Xmin,Ymin,Xmax,Ymax
      call release(X2)
      call release(Y2)
      call release(tab2)
  !--------------------------------------------
  !  ADD WCS KEYWORDS
  !--------------------------------------------
      call setAttribute(outset,"EQUINOX",2.000000E+03,"","Equinox for sky coordinate x/y axes")
      call setAttribute(outset,"RADECSYS","FK5","World coord. system for this file")
      call setAttribute(X,"TCTYP","RA---TAN","Coord. type: RA tangent plane projection")
        value = nint(imsize + 1)
      call setAttribute(X,"TCRPX",value,"","axis reference pixel of projected image")
      call setAttribute(X,"TCRVL",outranom*radian,"deg","coord. at X axis ref. pixel")
      call setAttribute(X,"TCDLT",-scaleX*radian,"deg/pix","X increment at ref. pixel")
      call setAttribute(X,"TLMIN",1,"","minimum legal QPOE projected image X axis value")
	value = nint(2*imsize)
      call setAttribute(X,"TLMAX",value,"","maximum legal QPOE projected image X axis value")
      call setAttribute(X,"TDMIN",Xmin,"","minimum projected image X axis data value")
      call setAttribute(X,"TDMAX",Xmax,"","maximum projected image X axis data value")
      call setAttribute(X,"TCUNI","deg","Physical units of X axis")
      call setAttribute(Y,"TCTYP","DEC--TAN","Coord. type: DEC tangent plane projection")
        value = nint(imsize + 1)
      call setAttribute(Y,"TCRPX",value,"","axis reference pixel of projected image")
      call setAttribute(Y,"TCRVL",outdecnom,"deg","coord. at Y axis ref. pixel")
      call setAttribute(Y,"TCDLT",scaleY*radian,"deg/pix","Y increment at ref. pixel")
      call setAttribute(Y,"TLMIN",1,"","minimum legal QPOE projected image Y axis value")
	value = nint(2*imsize)
      call setAttribute(Y,"TLMAX",value,"","maximum legal QPOE projected image Y axis value")
      call setAttribute(Y,"TDMIN",Ymin,"","minimum projected image Y axis data value")
      call setAttribute(Y,"TDMAX",Ymax,"","maximum projected image Y axis data value")
      call setAttribute(Y,"TCUNI","deg","Physical units of Y axis")
    else
      write (errstr,*) "X and/or Y columns do not exist in one or both input files - no reprojecting"
      call warning("noXYColumns")
    endif    
    call release(tab)

    end subroutine XY
!******************************************************************************
    subroutine antisp(dxn,dyn,dnomra,dnomde,dra1,dde1)

    implicit none    
    real(double)  :: dinra, dinde, dxn1, dyn1, dxn, dyn, dnomra, dnomde, &
                     dra1, dde1, d6, d1
    real, parameter               :: radian = 57.29577951E0, &
                                   skypixsize = 1.3888888888E-5, &!0.05arcsec, 
                                   pi = 3.1415926536E0
    dinra=dnomra
    dinde=dnomde
    dxn1=-dxn
    dyn1=-dyn
    d6 = pi
    d1 = 0.0
    call dtra(-dinde,d6,d1,dxn1,dyn1,dra1,dde1)
    dra1=dra1+dnomra
    if (dra1 > 2.0*pi) dra1 = dra1 - (2.0*pi)

    end subroutine antisp
!******************************************************************************
    subroutine sphere(ra,dec,ranom,sdecnom,cdecnom,relx,rely)

    implicit none
    real(double) :: diffra, ranom, sinra, cosra, sindec, cosdec, &
                    ra, dec, sdecnom, cdecnom, relx, rely, d1s, dh, d1c, &
                    ranew, d2s, d2c, decnew

    diffra = ra - ranom
    sinra = dsin(diffra)
    cosra = dcos(diffra)
    sindec = dsin(dec)
    cosdec = dcos(dec)
!       residual alpha ranew
    d1s = sinra * cosdec
    dh =  - cosdec * cosra
    d1c = sdecnom*sindec - dh*cdecnom
    ranew = datan2(d1s,d1c)
!       residual delta decnew
    d2s = sindec*cdecnom + dh*sdecnom 
    d2c = dsqrt(1.0d0 - d2s*d2s)
    decnew = datan2(d2s,d2c)
!       convert to x,y in NS-system
    relx = ranew
    rely = decnew

    end subroutine sphere
!******************************************************************************
    subroutine dtra(drd,d6,d8,d2,d1,d2n,d1n)

    implicit none    
    real(double)  :: d2, d6, d1, dr, drd, dsin2, dcos2, dsin1, dcos1, &
                     dsinr, dcosr, dx, d7, dy, d2n, d8, ddns, ddnc, d1n
    real, parameter               :: radian = 57.29577951E0, &
                                   skypixsize = 1.3888888888E-5, &!0.05arcsec, 
                                   pi = 3.1415926536E0, &
                                   df = 1.E0
      d2=d2-d6
      d1=d1*df
      d2=d2*df
      dr=drd*df
      dsin2=dsin(d2)   
      dcos2=dcos(d2)
      dsin1=dsin(d1)
      dcos1=dcos(d1)
      dsinr=dsin(dr)
      dcosr=dcos(dr)
      dx=-dsin2*dcos1
      d7=dcos1*dcos2
      dy=dsinr*dsin1-d7*dcosr
      d2n=datan2(dx,dy)
      d2n=d2n/df
      if (d2n < 0.D0) d2n=d2n+(2*pi)
      d2n=d8+d2n    
      if (d2n < 0.D0) d2n=d2n+(2*pi)
      if (d2n > 2.0*pi) d2n=d2n-(2*pi)
      ddns=d7*dsinr+dcosr*dsin1
      ddnc=dsqrt(1.D0-ddns*ddns)
      d1n=datan2(ddns,ddnc)
      d1n=d1n/df

    end subroutine dtra
!******************************************************************************
    subroutine hk
    end subroutine hk
!******************************************************************************
    subroutine orb(type)

  implicit none

  type(DataSetT)                :: set1, set2, outset
  type(TableT)                  :: tab1, tab2

  character(len=40)             :: attr1, attr2
  character(len=8)              :: type

  logical			:: reverse, firsttab
                                          
  integer, parameter            :: Ncolumns = 7
  character(len=8), parameter   :: tabname = "ORBIT   "
  character(len=8), dimension(7), parameter :: colname = &
    (/"TIME    ","X       ","Y       ","Z       ",& 
      "VX      ","VY      ","VZ      "/)
  character(len=3), dimension(7), parameter :: coltype = &
    (/"R64","R64","R64","R64","R64","R64","R64"/)

!------------------------------------------------
!	GET PARAMETERS 	
!------------------------------------------------
    set1 = dataset(stringParameter("set1"),READ)
    set2 = dataset(stringParameter("set2"),READ)
    tab1 = table(set1,"ORBIT")
    tab2 = table(set2,"ORBIT")
!------------------------------------------
!     Files consistent?
!------------------------------------------
    write (errstr,*) "Checking files: ",trim(stringParameter("set1"))," and ",trim(stringParameter("set2"))
    call message(AppMsg,Verbose)
    attr1 = stringAttribute(tab1,"TELESCOP")
    attr2 = stringAttribute(tab2,"TELESCOP")
    if (trim(attr1) /= trim(attr2)) then  
      call error("filesNonCompatible","Input files are non-compatible (TELESCOP)")
    endif
!    write (errstr,*) "Telescope: ",trim(attr1)
!    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"Files consistent")
!------------------------------------------
!     Merge orbit extensions 
!------------------------------------------
    firsttab = .TRUE.
    call mergetab(set1,set2,type,tabname,Ncolumns,colname,coltype,&
                  reverse,firsttab,outset)
!------------------------------------------
!     
!------------------------------------------

    call release(outset)

    end subroutine orb
!******************************************************************************
end module merge_mod

